import kaskada as kd
import pytest


@pytest.mark.skip("temporary skip, failing windows build")
async def test_read_parquet(golden) -> None:
    source = await kd.sources.Parquet.create(
        "../testdata/purchases/purchases_part1.parquet",
        time_column="purchase_time",
        key_column="customer_id",
    )
    golden.jsonl(source)

    await source.add_file("../testdata/purchases/purchases_part2.parquet")
    golden.jsonl(source)


@pytest.mark.skip("temporary skip, failing windows build")
async def test_read_parquet_with_subsort(golden) -> None:
    source = await kd.sources.Parquet.create(
        "../testdata/purchases/purchases_part1.parquet",
        time_column="purchase_time",
        key_column="customer_id",
        subsort_column="subsort_id",
    )
    golden.jsonl(source)

    await source.add_file("../testdata/purchases/purchases_part2.parquet")
    golden.jsonl(source)


# Verifies that we drain the output and progress channels correctly.
#
# When the parquet file contains more rows than
# (CHANNEL_SIZE / MAX_BATCH_SIZE), the channels previously filled
# up, causing the sender to block. This test verifies that the
# channels correctly drain, allowing the sender to continue.
# See https://github.com/kaskada-ai/kaskada/issues/775
@pytest.mark.skip("temporary skip, failing windows build")
async def test_large_parquet_file(golden) -> None:
    source = await kd.sources.Parquet.create(
        "../testdata/parquet/purchases_100k.parquet",
        time_column="time",
        key_column="user",
    )
    user = source.col("user")
    amount = source.col("amount")

    # Add a filter to reduce the output file size while ensuring the entire
    # file is still processed
    predicate = user.eq("5fec83d4-f5c6-4943-ab05-2b6760330daf").and_(amount.gt(490))
    golden.jsonl(source.filter(predicate))


@pytest.mark.skip("temporary skip, failing windows build")
async def test_time_column_as_float_can_cast_s(golden) -> None:
    source = await kd.sources.Parquet.create(
        "../testdata/purchases/purchases_float.parquet",
        time_column="purchase_time",
        key_column="customer_id",
        time_unit="s",
    )
    golden.jsonl(source)


@pytest.mark.skip("temporary skip, failing windows build")
async def test_time_column_as_float_can_cast_ns(golden) -> None:
    source = await kd.sources.Parquet.create(
        "../testdata/purchases/purchases_float.parquet",
        time_column="purchase_time",
        key_column="customer_id",
    )
    golden.jsonl(source)


@pytest.mark.skip("temporary skip, failing windows build")
async def test_with_space_in_path(golden) -> None:
    source = await kd.sources.Parquet.create(
        "../testdata/purchases/purchases part1.parquet",
        time_column="purchase_time",
        key_column="customer_id",
        subsort_column="subsort_id",
    )
    golden.jsonl(source)


@pytest.mark.skip("temporary skip, failing windows build")
async def test_with_trailing_slash(golden) -> None:
    source = await kd.sources.Parquet.create(
        "../testdata/purchases/purchases part1.parquet/",
        time_column="purchase_time",
        key_column="customer_id",
        subsort_column="subsort_id",
    )
    golden.jsonl(source)


@pytest.mark.svc("minio")
async def test_read_parquet_from_s3_minio(minio, golden) -> None:
    # Upload a parquet file to minio for testing purposes
    (minio_host, minio_port) = minio
    import boto3

    aws_endpoint = f"http://{minio_host}:{minio_port}"
    # Defaults set in the `pytest-docker-fixtures`
    # https://github.com/guillotinaweb/pytest-docker-fixtures/blob/236fdc1b6a9db03640040af2baf3bd3dfcc8d187/pytest_docker_fixtures/images.py#L42
    aws_access_key_id = "x" * 10
    aws_secret_access_key = "x" * 10
    s3_client = boto3.client(
        "s3",
        endpoint_url=aws_endpoint,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        # aws_session_token=None,
        # config=boto3.session.Config(signature_version='s3v4'),
        # verify=False
    )
    s3_client.create_bucket(Bucket="test-bucket")
    s3_client.upload_file(
        "../testdata/purchases/purchases_part1.parquet",
        "test-bucket",
        "purchases_part1.parquet",
    )

    # This is a hack.
    # The session (which is only created once) will cache the S3 client.
    # This means that generally, changing the environment variables would
    # require creating a new session to be picked up.
    # Currently, we don't allow recreating the session (or scoping it).
    # We probably should.
    #
    # But... this still works. The trick is that the S3 client isn't
    # created until the first S3 URL is used. As long as we set the
    # environment variables before that, they will be picked up and
    # stored in the session appropriately.
    import os

    os.environ["AWS_ENDPOINT"] = aws_endpoint
    os.environ["AWS_ACCESS_KEY_ID"] = aws_access_key_id
    os.environ["AWS_SECRET_ACCESS_KEY"] = aws_secret_access_key
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
    # Minio requires HTTP.
    os.environ["AWS_ALLOW_HTTP"] = "true"

    source = await kd.sources.Parquet.create(
        "s3://test-bucket/purchases_part1.parquet",
        time_column="purchase_time",
        key_column="customer_id",
    )
    golden.jsonl(source)



#!/usr/bin/env python
#
# Reddit Kaskada Consumer
#
# This script demonstrates the use of Kaskada to consume and compute over
# Reddit updates.

import asyncio
import os
import time

import asyncpraw
import kaskada as kd
import pyarrow as pa


async def main():
    # Initialize the Kaskada session so we can use it for stream processing
    kd.init_session()

    # Create the Reddit client.
    reddit = asyncpraw.Reddit(
        client_id=os.getenv("REDDIT_CLIENT_ID"),
        client_secret=os.getenv("REDDIT_CLIENT_SECRET"),
        user_agent=os.getenv("REDDIT_USER_AGENT", "kaskada-demo"),
    )

    # [start_setup]
    # Setup the data source.
    # This defintes (most of) the schema of the events we'll receive,
    # and tells Kaskada which fields to use for time and initial entity.
    #
    # We'll push events into this source as they arrive in real-time.
    comments = kd.sources.PyDict(
        schema=pa.schema(
            [
                pa.field("author", pa.string()),
                pa.field("body", pa.string()),
                pa.field("permalink", pa.string()),
                pa.field("submission_id", pa.string()),
                pa.field("subreddit", pa.string()),
                pa.field("ts", pa.float64()),
            ]
        ),
        time_column="ts",
        key_column="submission_id",
        time_unit="s",
    )
    # [end_setup]

    # [start_incoming]
    # Handler to receive new comments as they're created
    async def receive_comments():
        # Creat the subreddit handle
        sr = await reddit.subreddit(os.getenv("SUBREDDIT", "all"))

        # Consume the stream of new comments
        async for comment in sr.stream.comments():
            # Add each comment to the Kaskada data source
            await comments.add_rows(
                {
                    "author": comment.author.name,
                    "body": comment.body,
                    "permalink": comment.permalink,
                    "submission_id": comment.submission.id,
                    "subreddit_id": comment.subreddit.display_name,
                    "ts": time.time(),
                }
            )

    # [end_incoming]

    # [start_result]
    # Handler for values emitted by Kaskada.
    async def receive_outputs():
        # We'll perform a very simple aggregation - key by author and count.
        comments_by_author = comments.with_key(comments.col("author"))

        # Consume outputs as they're generated and print to STDOUT.
        async for row in comments_by_author.count().run_iter(kind="row", mode="live"):
            print(f"{row['_key']} has posted {row['result']} times since startup")

    # [end_result]

    # [start_run]
    # Kickoff the two async processes concurrently.
    await asyncio.gather(receive_comments(), receive_outputs())
    # [end_run]


if __name__ == "__main__":
    asyncio.run(main())



import kaskada as kd
import pandas as pd


async def test_float_seconds(golden) -> None:
    data = {"time": [1671477472.026119], "user": ["tom"]}
    df = pd.DataFrame(data)
    table = await kd.sources.Pandas.create(
        df, time_column="time", key_column="user", time_unit="s"
    )

    golden.jsonl(table)


async def test_float_milliseconds(golden) -> None:
    data = {"time": [1671477472026.119], "user": ["tom"]}
    df = pd.DataFrame(data)
    table = await kd.sources.Pandas.create(
        df, time_column="time", key_column="user", time_unit="ms"
    )

    golden.jsonl(table)


async def test_float_nanoseconds(golden) -> None:
    data = {"time": [1671477472026119000], "user": ["tom"]}
    df = pd.DataFrame(data)
    table = await kd.sources.Pandas.create(
        df, time_column="time", key_column="user", time_unit="ns"
    )

    golden.jsonl(table)


async def test_add_dataframe(golden) -> None:
    df1 = pd.DataFrame(
        {
            "time": [1000000000000],
            "key": ["a"],
        }
    )

    table = await kd.sources.Pandas.create(df1, time_column="time", key_column="key")
    golden.jsonl(table)

    df2 = pd.DataFrame(
        {
            "time": [1000000000000],
            "key": ["a"],
        }
    )
    await table.add_data(df2)
    golden.jsonl(table)


async def test_add_dataframe_with_subsort(golden) -> None:
    df1 = pd.DataFrame(
        {
            "time": [1000000000000],
            "key": ["a"],
            "subsort": [1],
        }
    )

    table = await kd.sources.Pandas.create(
        df1,
        time_column="time",
        key_column="key",
        subsort_column="subsort",
    )
    golden.jsonl(table)

    df2 = pd.DataFrame(
        {
            "time": [1000000000000],
            "key": ["a"],
            "subsort": [2],
        }
    )
    await table.add_data(df2)
    golden.jsonl(table)



from datetime import timedelta

import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1997-01-18T16:40:00,A,,9",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


@pytest.fixture(scope="module")
async def filter_source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "thread_ts,ts,channel",
            "null,1691762610.0,Project",
            "null,1691762620.0,Project",
            "null,1691762630.0,Project",
            "1691762650.0,1691762650.0,Project",
            "1691762650.0,1691762660.0,Project",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="ts", key_column="channel", time_unit="s"
    )


async def test_shift_by_timedelta(source, golden) -> None:
    time = source.col("time")
    golden.jsonl(
        kd.record(
            {
                "time": time,
                "shift_by_1_s": time.shift_by(timedelta(seconds=1)),
                "shift_by_1_m": time.shift_by(timedelta(minutes=1)),
            }
        )
    )


async def test_shift_collect(source, golden) -> None:
    golden.jsonl(
        source.record(
            lambda input: {
                "time": input.col("time"),
                "ms": input.col("m").collect(max=10),
                "m": input.col("m"),
            }
        )
        # Currently, the Pandas comparison method being used doesn't handle
        # date-time like fields nested within a list. So we expand things out.
        #
        # TODO: Improve the golden testing so this isn't necessary.
        .extend(
            lambda base: {
                "shift_by_1_s_time": base.shift_by(timedelta(seconds=1)).col("time"),
                "shift_by_1_s_ms": base.shift_by(timedelta(seconds=1)).col("ms"),
                "shift_by_1_m_time": base.shift_by(timedelta(minutes=1)).col("time"),
                "shift_by_1_m_ms": base.shift_by(timedelta(minutes=1)).col("ms"),
            }
        )
    )


# Regression test for https://github.com/kaskada-ai/kaskada/issues/726
async def test_filter_to_shift(filter_source, golden) -> None:
    messages = filter_source.filter(filter_source.col("thread_ts").is_null())
    messages = messages.shift_by(timedelta(seconds=5))
    golden.jsonl(messages)



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n,o",
            "1996-12-19T16:39:57,A,5,10,15",
            "1996-12-19T16:39:58,B,24,3,15",
            "1996-12-19T16:39:59,A,17,6,15",
            "1996-12-19T16:40:00,A,,9,15",
            "1996-12-19T16:40:01,A,12,,15",
            "1996-12-19T16:40:02,A,,,15",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_coalesce(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    golden.jsonl(kd.record({"m": m, "n": n, "coalesced_val": m.coalesce(n)}))


async def test_coalesce_three(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    o = source.col("o")
    golden.jsonl(kd.record({"m": m, "n": n, "o": o, "coalesced_val": m.coalesce(n, o)}))



"""Control which points and entities are output."""

from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class History:
    """Execution options for queries producing all historic points."""

    #: If set, only returns points after this time.
    #:
    #: Setting this allows incremental execution to use a checkpoint
    #: from a time before the `since` time.
    since: Optional[datetime] = None

    #: Only return points less than or equal to this time.
    #: If not set, the current time will be used.
    until: Optional[datetime] = None


@dataclass
class Snapshot:
    """Execution options for queries producing snapshots at a specific time."""

    #: If set, only includes entities that changed after this time.
    #:
    #: Snapshot queries support incremental execution even when this isn't set.
    #: However, every snapshot will include every entity unless this is set.
    #: When writing results to an external store that already has values
    #: from an earlier snapshot, this can be used to reduce the amount of
    #: data to be written.
    changed_since: Optional[datetime] = None

    #: If set, produces the snapshot at the given time.
    #: If not set, the current time will be used.
    at: Optional[datetime] = None



import kaskada as kd
import pyarrow as pa
import pytest


@pytest.fixture
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n,t",
            "1996-12-19T16:39:57,A,5,10,1996-12-19T16:42:57",
            "1996-12-19T16:39:58,B,24,3,1996-12-19T16:39:59",
            "1996-12-19T16:39:59,A,17,6,",
            "1996-12-19T16:40:00,A,,9,1996-12-19T16:41:00",
            "1996-12-19T16:40:01,A,12,,1996-12-19T16:42:01",
            "1996-12-19T16:40:02,A,,,1996-12-19T16:43:02",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_seconds_since_previous(golden, source) -> None:
    t = source.col("time")
    golden.jsonl(
        kd.record(
            {
                "seconds_since": t.seconds_since_previous().cast(pa.int64()),
                "seconds_since_1": t.seconds_since_previous(1).cast(pa.int64()),
                "seconds_since_2": t.seconds_since_previous(2).cast(pa.int64()),
            }
        )
    )



"""Provide the base-class for Kaskada sources."""
import os
from typing import Literal, Optional

import kaskada._ffi as _ffi
import pyarrow as pa
from typing_extensions import TypeAlias

from .._session import _get_session
from .._timestream import Timestream


_TABLE_NUM: int = 0

TimeUnit: TypeAlias = Literal["ns", "us", "ms", "s"]


class Source(Timestream):
    """A source (input) Timestream."""

    # TODO: Clean-up naming on the FFI side.
    _ffi_table: _ffi.Table

    def __init__(
        self,
        schema: pa.Schema,
        time_column: str,
        key_column: str,
        queryable: bool = True,
        subsort_column: Optional[str] = None,
        grouping_name: Optional[str] = None,
        time_unit: Optional[TimeUnit] = None,
        source: Optional[str] = None,
    ):
        """Create a new source."""
        assert isinstance(schema, pa.Schema)

        # Fix the schema. The fields should be non-nullable.
        def fix_field(field: pa.Field) -> pa.Field:
            if field.name in [
                time_column,
                key_column,
                subsort_column,
            ]:
                field = field.with_nullable(False)
            if isinstance(field.type, pa.TimestampType):
                field = field.with_type(pa.timestamp(field.type.unit, tz=None))
            return field

        fields = [fix_field(f) for f in schema]
        schema = pa.schema(fields)

        Source._validate_column(time_column, schema)
        Source._validate_column(key_column, schema)
        Source._validate_column(subsort_column, schema)

        # Hack -- Sparrow currently requires tables be named.
        global _TABLE_NUM
        name = f"table{_TABLE_NUM}"
        _TABLE_NUM += 1

        ffi_table = _ffi.Table(
            _get_session(),
            name,
            time_column,
            key_column,
            schema,
            queryable,
            subsort_column,
            grouping_name,
            time_unit,
            source,
        )
        super().__init__(ffi_table.expr())
        self._schema = schema
        self._ffi_table = ffi_table

    @property
    def name(self) -> str:
        """Get the current table name."""
        return self._ffi_table.name

    # TODO: Most of these checks exist in Sparrow. We should just surface
    # those errors more cleanly.
    @staticmethod
    def _validate_column(field_name: Optional[str], schema: pa.Schema) -> None:
        if field_name is not None:
            field = schema.field(field_name)
            if field is None:
                raise KeyError(f"Column {field_name!r} does not exist")
            if field.nullable:
                raise ValueError(f"Column: {field_name!r} must be non-nullable")

    @staticmethod
    def _get_absolute_path(path: Optional[str]) -> Optional[str]:
        if path is None or os.path.isabs(path):
            return path
        return os.path.abspath(path)



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def string_source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "2021-01-01T00:00:00,Ben,hello,",
            "2021-01-01T00:00:00,Ryan,,",
            "2021-01-02T00:00:00,Ryan,world,",
            "2021-01-03T00:00:00,Ben,hi,",
            "2021-01-04T00:00:00,Ben, ,",
            "2021-01-04T00:00:00,Ryan,earth,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_hash_string(string_source, golden) -> None:
    m = string_source.col("m")
    golden.jsonl(kd.record({"m": m, "hash_m": m.hash()}))


@pytest.fixture(scope="module")
async def integer_source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "2021-01-01T00:00:00,Ben,5,",
            "2021-01-01T00:00:00,Ryan,8,",
            "2021-01-02T00:00:00,Ryan,9,",
            "2021-01-03T00:00:00,Ben,8,",
            "2021-01-04T00:00:00,Ben,,",
            "2021-01-04T00:00:00,Ryan,9,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_hash_integer(integer_source, golden) -> None:
    m = integer_source.col("m")
    golden.jsonl(kd.record({"m": m, "hash_m": m.hash()}))



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1996-12-19T16:40:00,A,,9",
            "1996-12-19T16:40:01,A,12,",
            "1996-12-19T16:40:02,A,,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_filter(source, golden) -> None:
    m = source.col("m")
    n = source.col("n")
    condition_m = m > 15
    condition_n = n > 5
    golden.jsonl(
        kd.record(
            {
                "m": m,
                "condition_m": condition_m,
                "filter_m": m.filter(condition_m),
                "n": n,
                "condition_n": condition_n,
                "filter_n": n.filter(condition_n),
            }
        )
    )


async def test_if_filter_n(source, golden) -> None:
    n = source.col("n")
    condition_n = n > 5
    golden.jsonl(
        kd.record(
            {
                "filter_n": n.if_(condition_n).filter(condition_n),
            }
        )
    )


# Regression test for https://github.com/kaskada-ai/kaskada/issues/766
async def test_filter_to_merge_preserves_interpolation(source, golden) -> None:
    n = source.col("n")
    predicate = n < 9
    golden.jsonl(
        kd.record(
            {"n": n, "predicate": predicate, "filter_sum": n.sum().filter(predicate)}
        )
    )



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,s,n,t",
            '1996-12-19T16:39:57,A,"hEllo",0,"hEllo"',
            '1996-12-19T16:39:58,B,"World",5,"world"',
            '1996-12-19T16:39:59,B,"hello world",-2,"hello world"',
            '1996-12-19T16:40:00,B,,-2,"greetings"',
            '1996-12-19T16:40:01,B,,2,"salutations"',
            '1996-12-19T16:40:02,B,"goodbye",,',
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_len(source, golden) -> None:
    s = source.col("s")
    golden.jsonl(kd.record({"len": s.len()}))



from datetime import datetime
from typing import AsyncIterator, Iterator

import kaskada as kd
import pandas as pd
import pytest


@pytest.fixture
async def source_int64() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m,n",
            "1996-12-19T16:39:57,A,5,10",
            "1996-12-19T16:39:58,B,24,3",
            "1996-12-19T16:39:59,A,17,6",
            "1996-12-19T16:40:00,A,,9",
            "1996-12-19T16:40:01,A,12,",
            "1996-12-19T16:40:02,A,,",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_iter_pandas(golden, source_int64) -> None:
    batches = source_int64.run_iter(row_limit=4, max_batch_size=2)

    # 4 rows, max 2 per batch = 2 batches
    golden.jsonl(next(batches))
    golden.jsonl(next(batches))
    with pytest.raises(StopIteration):
        next(batches)


async def test_iter_rows(source_int64) -> None:
    results: Iterator[dict] = source_int64.run_iter("row", row_limit=2)
    assert next(results)["m"] == 5
    assert next(results)["m"] == 24
    with pytest.raises(StopIteration):
        next(results)


@kd.udf("add_one<N: number>(x: N) -> N")
def add_one(x: pd.Series) -> pd.Series:
    """Use Pandas to add one."""
    return x + 1


async def test_iter_udf(source_int64) -> None:
    results = source_int64.col("m").pipe(add_one).run_iter("row", row_limit=2)
    assert next(results)["result"] == 6
    assert next(results)["result"] == 25


async def test_iter_pandas_async(golden, source_int64) -> None:
    batches: AsyncIterator[pd.DataFrame] = source_int64.run_iter(
        row_limit=4, max_batch_size=2
    )

    # 4 rows, max 2 per batch = 2 batches.

    # We could test using `await anext(batches)`, but that wasn't introduced
    # until Python 3.10. Since everything else works in 3.8 and 3.9, we just
    # call `__anext__` directly.
    golden.jsonl(await batches.__anext__())
    golden.jsonl(await batches.__anext__())
    with pytest.raises(StopAsyncIteration):
        await batches.__anext__()


async def test_iter_pandas_async_live(golden, source_int64) -> None:
    data2 = "\n".join(
        [
            "time,key,m,n",
            "1996-12-20T16:39:57,A,5,10",
            "1996-12-20T16:39:58,B,24,3",
            "1996-12-20T16:39:59,A,17,6",
            "1996-12-20T16:40:00,C,,9",
            "1996-12-20T16:40:01,A,12,",
            "1996-12-20T16:40:02,A,,",
        ]
    )

    execution = source_int64.run_iter(mode="live")

    # Await the first batch.
    golden.jsonl(await execution.__anext__())

    # Add data and await the second batch.
    await source_int64.add_string(data2)
    golden.jsonl(await execution.__anext__())

    execution.stop()
    with pytest.raises(StopAsyncIteration):
        print(await execution.__anext__())


async def test_snapshot(golden, source_int64) -> None:
    query = source_int64.col("m").sum()
    golden.jsonl(query.to_pandas(kd.results.Snapshot()))
    golden.jsonl(
        query.to_pandas(
            kd.results.Snapshot(
                changed_since=datetime.fromisoformat("1996-12-19T16:39:59+00:00")
            )
        )
    )
    golden.jsonl(
        query.to_pandas(
            kd.results.Snapshot(at=datetime.fromisoformat("1996-12-20T12:00:00+00:00"))
        )
    )


async def test_history(golden, source_int64) -> None:
    query = source_int64.col("m").sum()
    golden.jsonl(query.to_pandas(kd.results.History()))
    golden.jsonl(
        query.to_pandas(
            kd.results.History(
                since=datetime.fromisoformat("1996-12-19T16:39:59+00:00")
            )
        )
    )
    golden.jsonl(
        query.to_pandas(
            kd.results.History(
                until=datetime.fromisoformat("1996-12-20T12:00:00+00:00")
            )
        )
    )
    golden.jsonl(
        query.to_pandas(
            kd.results.History(
                since=datetime.fromisoformat("1996-12-19T16:39:59+00:00"),
                until=datetime.fromisoformat("1996-12-20T12:00:00+00:00"),
            )
        )
    )



import kaskada as kd
import pyarrow as pa
import pytest


async def test_read_pydict(golden) -> None:
    source = await kd.sources.PyDict.create(
        rows=[{"time": "1996-12-19T16:39:57", "user": "A", "m": 5, "n": 10}],
        time_column="time",
        key_column="user",
    )
    golden.jsonl(source)

    await source.add_rows(
        [
            {"time": "1996-12-19T16:40:57", "user": "A", "m": 8, "n": 10},
            {"time": "1996-12-19T16:41:57", "user": "B", "m": 5},
        ]
    )

    golden.jsonl(source)
    await source.add_rows({"time": "1996-12-19T16:42:57", "user": "A", "m": 8, "n": 10})
    golden.jsonl(source)


async def test_read_pydict_lists(golden) -> None:
    source = await kd.sources.PyDict.create(
        rows=[{"time": "1996-12-19T16:39:57", "user": "A", "m": [5, 10], "n": 10}],
        time_column="time",
        key_column="user",
    )
    golden.jsonl(source)

    await source.add_rows(
        [
            {"time": "1996-12-19T16:40:57", "user": "A", "m": [], "n": 10},
            {"time": "1996-12-19T16:41:57", "user": "A", "n": 10},
        ]
    )
    golden.jsonl(source)


async def test_read_pydict_ignore_column(golden) -> None:
    # Schema is determined from first row, and doesn't contain an "m" column.
    source = await kd.sources.PyDict.create(
        rows=[{"time": "1996-12-19T16:39:57", "user": "A", "n": 10}],
        time_column="time",
        key_column="user",
    )
    golden.jsonl(source)

    await source.add_rows(
        [
            {"time": "1996-12-19T16:40:57", "user": "A", "m": 83, "n": 10},
            {"time": "1996-12-19T16:41:57", "user": "A", "m": 12},
        ]
    )
    golden.jsonl(source)


async def test_read_pydict_empty() -> None:
    source = kd.sources.PyDict(
        schema=pa.schema(
            [
                ("time", pa.timestamp("ns")),
                ("user", pa.string()),
                ("x", pa.int64()),
            ]
        ),
        time_column="time",
        key_column="user",
    )

    result = source.preview()

    assert result.empty
    assert list(result.columns) == ["_time", "_key", "time", "user", "x"]
    assert result["x"].dtype == "int64"


async def test_read_pydict_empty_live(golden) -> None:
    source = kd.sources.PyDict(
        schema=pa.schema(
            [
                ("time", pa.string()),
                ("user", pa.string()),
                ("x", pa.int64()),
            ]
        ),
        time_column="time",
        key_column="user",
    )

    execution = source.run_iter(mode="live")
    await source.add_rows({"time": "1996-12-19T16:39:57Z", "user": "A", "x": 5})
    golden.jsonl(await execution.__anext__())

    execution.stop()
    with pytest.raises(StopAsyncIteration):
        print(await execution.__anext__())



import kaskada as kd
import pytest


@pytest.fixture(scope="module")
async def source() -> kd.sources.CsvString:
    content = "\n".join(
        [
            "time,key,m",
            "2021-01-01T00:00:00,Ben,Hello World",
            "2021-01-02T00:00:00,Ryan,",
            "2021-01-02T00:00:00,Ryan,Hi Earth",
            "2021-01-03T00:00:00,Ben,Hello",
            "2021-01-03T00:00:00,Ben,",
            "2021-01-04T00:00:00,Ryan,hi",
        ]
    )
    return await kd.sources.CsvString.create(
        content, time_column="time", key_column="key"
    )


async def test_upper(source, golden) -> None:
    m = source.col("m")
    golden.jsonl(kd.record({"m": m, "upper_m": m.upper()}))
